//////////////////////////////////////////////////////////////////////////////////////////////
//
//  Created by Daniel Pantea, from GE Healthcare.
//  No guaranties given. Use on your own risk!
//  You may copy, modify and do what ever you want with this code.
//
//  Uses <SlickEdit-dir>\win\SEtoVS.dll (95% inspired from Gary Ash's original idea: SlickEditIntegration.dll).
// 
// date version: 2013-Aug-13
// 
//////////////////////////////////////////////////////////////////////////////////////////////


#pragma option(strict, on)
#include "slick.sh"


#define MY_MACRO_DEBUG  0
#if MY_MACRO_DEBUG != 0
/* for debug only */
    #define my_say(x)  say((x))
#else
/* nothing */
    #define my_say(x)
#endif


// done - count TABS in current line
// done - add breakpoint not at current line, but where VC saiz: actually get all breakpoints after toggled current line, not before!
// done - delete all breakpoints if VS isn't available!
// done - when come back from VC, SE asks to open VC (even it is not started)
// done - use full path when search for workspace file, to avoid geting the first from multiple reasults, like "Alarms.cpp"
// done - timer callbacks: use reference for timer-ID! Not param and return! Avoid timers?
// done - ask user for keybindings
// done - update breakpoints when SE gets focus
// done - always call my_VS_UpdateBreaks via _post_call! 
// done - see Maybe_init_all
// done - updated for 2010
// todo - ...
//        Add more comments (check spelling), especially to function headers!


// launches the Visual Studio .NET
extern void _dll_LaunchVsIde();
// Start the Visual Studio debugger
extern void _dll_VsStartDebug();
// automate VS to lookup help on the word at the current cursor position
extern void _dll_VsHelp();
// Clear all Visual Studio break points
extern void _dll_ClearBreakpoints();
// Toggle a Visual Studio break point on and off
extern void _dll_ToggleBreakpoint();
// Sync Visual Studio with current possition from Slick Edit
extern void _dll_SetVsWhereSeIs();
// set when sync from VS to SE, because SE's current bufer would give previous file's name and not the one we just switched (also silent connect to VS)
extern void _dll_SetKickBack();
// Get nr of breakpoints from current file into Slick Edit
extern int _dll_Brk_GetCnt();
// reset the index of evaluated breakpoint
extern void _dll_Brk_ResetEvalNdx();
// go to next breakpoint which will be evaluated
extern void _dll_Brk_IncEvalNdx();
// get the position of evaluated breakpoint
extern long _dll_Brk_GetPos();
// get the status of evaluated breakpoint
extern int _dll_Brk_GetState();
// is IDE up and connected? Don't try to connect if not!
extern int _dll_IsVsIdeUp();
// set when dll was loaded succesfully
static boolean gIsDllLoaded = false;

/* my default keys:
def 'A-S' = my_VS_SyncToSlickEdit;
def 'A-F9' = my_VS_ClearAllBreaks;
def 'F9' = my_VS_ToggleBreak;
def 'F5' = my_VS_GoDebuging;
*/



struct _VCppData{
   //int gMyUpdateBreaksTimerHandler;       // handler of the timer which updates the breakpoints
   //boolean gTempDisableSwitchBuffUpdate;  // temporary disable breakpoint-updates triggered by buffer-changing
   _str gOldBuffName;                     // used to detect when a new buffer was changed
   int _pic_break_enabled_marker;         // handler to breakpoint-enabled picture
   int gpictype_break_enabled;            // the allocated type for the breakpoint-enabled marker
   int _pic_break_disabled_marker;        // handler to breakpoint-disabled picture
   int gpictype_break_disabled;           // the allocated type for the breakpoint-disabled marker
};

static struct _VCppData VCppData;


// exported for manual dll reloading
_command boolean my_VS_LoadDll() name_info (','VSARG2_MACRO|VSARG2_READ_ONLY)
{
   _str name = "SEtoVS.dll";
   // compute destination file name
   _str root_dir=get_env('VSROOT');
   _maybe_append_filesep(root_dir);
   _str dest = root_dir :+ "win" :+ FILESEP :+ name;
   int status = dload(maybe_quote_filename(dest));
   if (status) {
      _str line = "Failed to load DLL: " :+ dest;
      _message_box(line);
      return false;
   }
   else {
      gIsDllLoaded = true;
      return true;
   }
}


// go to Visual Studio
_command void my_VS_SyncToSlickEdit() name_info (','VSARG2_MACRO|VSARG2_READ_ONLY|VSARG2_REQUIRES_EDITORCTL)
{
   //popup_message/*my_say*/("my_VS_GotoThere");
   _dll_SetVsWhereSeIs();
}


// start VisualStudio debugger
_command void my_VS_GoDebuging() name_info (','VSARG2_MACRO|VSARG2_READ_ONLY|VSARG2_REQUIRES_EDITORCTL)
{
   _dll_VsStartDebug();
}


// lunch VisualStudio IDE
_command void my_VS_StartIDE() name_info (','VSARG2_MACRO|VSARG2_READ_ONLY|VSARG2_REQUIRES_EDITORCTL)
{
   _dll_LaunchVsIde();
}


#if 0
// fwd declaration
void my_VS_UpdateBreaksDelayed();

// called externally for syncronisation
// 
// ex:         from Visual C++ 2003, External Tools:
// Title:      SlickEdit synchro
// Command:    C:\Program Files\SlickEditV18.0.1\win\vs.exe
// Arguments:  "-# SetSeWhereVsIs $(ItemPath)|$(CurLine)|$(CurCol)""
// 
_command int SetSeWhereVsIs(...) name_info(','VSARG2_CMDLINE|VSARG2_MARK|VSARG2_READ_ONLY)
{
   //message("SetSeWhereVsIs");
   typeless nLineNr = 0;   // line number where should we go
   typeless nColNr = 0;    // column position
   _str strFileName = '';  // filename
   _str param = arg(1);    // get the params
   if (param=="") {
      message('Please specify file name and line number'); delay(10);
      return -1;
   }
   //popup_message (param);
   parse param with strFileName '|' nLineNr '|' nColNr;    // split the params
   //popup_message ('"'strFileName'" - "'nLineNr'" - "'nColNr'"');
   strFileName = strip(strFileName, 'B');          // remove spaces

   // replacement for _WorkspaceFindFile
   _str strFilePath = my_WorkspaceFindFile(strFileName);
   if(strFilePath != "")
   {
      VCppData.gTempDisableSwitchBuffUpdate = true;
      my_say("SetSeWhereVsIs sets gTempDisableSwitchBuffUpdate");
      status := edit(maybe_quote_filename(strFilePath), EDIT_DEFAULT_FLAGS);
      if (status == 0) {
         goto_line(nLineNr); //p_RLine = nLineNr;
         p_col=nColNr;
         message('sync to "'strFileName'" - "'nLineNr'" - "'nColNr'"');
         _dll_SetKickBack(); // extract breakpoints using VS's current document, not SE's current buffer
         delay(2);
         //popup_message("called my_VS_UpdateBreaks, now exiting my_ESP_WksGoTo");
         if (VCppData.gMyUpdateBreaksTimerHandler <= -1) {
            VCppData.gMyUpdateBreaksTimerHandler = _set_timer(200, my_VS_UpdateBreaksDelayed);
         }
         //popup_message("set the timer");
         //popup_message("called my_VS_UpdateBreaks, now exiting SetSeWhereVsIs");
      }
      my_say("SetSeWhereVsIs resets gTempDisableSwitchBuffUpdate");
      VCppData.gTempDisableSwitchBuffUpdate = false;
   }
   else
   {
      sticky_message ('Can''t go to "'strFileName'" - "'nLineNr'" - "'nColNr'"!');
   }
   return 0;
}


// inspired from "files-workspace" toobar
_str my_WorkspaceFindFile(_str strFullFileName = "")
{
   //popup_message/*my_say*/("my_WorkspaceFindFile");
   _str strRet = "";
   _str strFileName = strip_filename(strFullFileName,'P');  // remove path
   if (strFileName != "") {
      // get the list of projects in this workspace
      _str projects[];
      projects._makeempty();
      _GetWorkspaceFiles(_workspace_filename, projects);
      for (i := 0; i < projects._length(); i++) {
         if (projects[i] == "") continue;
         // get the active project handle
         _str fullPath = _AbsoluteToWorkspace(projects[i]);
         int hProject = _ProjectHandle(fullPath);
         if (hProject > 0) {
            // get the list of files in this project
            _str fileList[];
            fileList._makeempty();
            getProjectFileList(fullPath, fileList);
            for (j := 0; j < fileList._length(); j++) {
               _str FullFileName = strip(fileList[j], 'B'); // remove spaces
               _str FileName = strip_filename(FullFileName,'P'); // remove path
               if (lowcase(strFileName) == lowcase(FileName)) {
                   if (lowcase(strFullFileName) == lowcase(FullFileName)) {
                      strRet = fileList[j];
                      //_message_box('fullPath = 'strRet);
                      break;
                   } else{
                      //_message_box('strFullFileName = "'strFullFileName'" FullFileName = "'FullFileName'"!');
                   }
               }
            }
         }
      }
      if (strRet != "") {
         // TODO: if the requested file isn't a workspace file
         // then maybe we can search for an existing file/buffer
         // and just open it, but for the momment not necessary
      }
   }
   return strRet;
}
#endif 

// internal use for loading some pics used for drawing breakpoints
//  (it does its job at first call, then just verify if everything's ok)
boolean my_VS_UpdateUsedBrkBitmaps()
{
   //popup_message/*my_say*/("my_VS_UpdateUsedBrkBitmaps");
   if (VCppData._pic_break_enabled_marker <= 0) {
#if __VERSION__ < 17.0
      VCppData._pic_break_enabled_marker = _update_picture(-1, "_breakpt.bmp");
#else
      //VCppData._pic_break_enabled_marker = _update_picture(-1, "_breakpt.ico");
      VCppData._pic_break_enabled_marker = _update_picture(-1, "_f_overlay_modified.svg");
#endif
      if (!VCppData._pic_break_enabled_marker) {
         message("VCppData._pic_break_enabled_marker = FALSE!");
         return false;
      }
   }
   if (VCppData.gpictype_break_enabled <= 0) {
      VCppData.gpictype_break_enabled = _MarkerTypeAlloc();
      if (!VCppData.gpictype_break_enabled) {
         message("VCppData.gpictype_break_enabled = FALSE!");
         return false;
      }
      else{
         //_MarkerTypeSetFlags(VCppData.gpictype_break_enabled, VSMARKERTYPEFLAG_DRAW_SQUIGGLY|VSMARKERTYPEFLAG_AUTO_REMOVE);
      }
   }
   if (VCppData._pic_break_disabled_marker <= 0) {
#if __VERSION__ < 17.0
      VCppData._pic_break_disabled_marker = _update_picture(-1, "_breakpn.bmp");
#else
      //VCppData._pic_break_disabled_marker = _update_picture(-1, "_breakpn.ico");
      VCppData._pic_break_disabled_marker = _update_picture(-1, "_f_overlay_new.svg");
#endif
      if (!VCppData._pic_break_disabled_marker) {
         message("VCppData._pic_break_disabled_marker = FALSE!");
         return false;
      }
   }
   if (VCppData.gpictype_break_disabled <= 0) {
      VCppData.gpictype_break_disabled = _MarkerTypeAlloc();
      if (!VCppData.gpictype_break_disabled) {
         message("VCppData.gpictype_break_disabled = FALSE!");
         return false;
      }
      else{
         //_MarkerTypeSetFlags(VCppData.gpictype_break_disabled, VSMARKERTYPEFLAG_DRAW_SQUIGGLY|VSMARKERTYPEFLAG_AUTO_REMOVE);
      }
   }
   //popup_message/*my_say*/("my_VS_UpdateUsedBrkBitmaps OK!");
   return true;
}

// retrieve breakpoints from VS and set them in current file from SE
static void my_VS_UpdateBreaks(int wid = _mdi.p_child.p_window_id)
{
   my_say("my_VS_UpdateBreaks");
   if(!_no_child_windows()) {
      // don't ask to start Visual Studio if not started!
      // because this function is called very often when buffer-changing!
      if (!_dll_IsVsIdeUp()) {
         my_say("my_VS_UpdateBreaks exits because NOT _dll_IsVsIdeUp");
         if (my_VS_UpdateUsedBrkBitmaps()) {
            //popup_message/*message*/("Clearing all breakpoint markers..." wid.p_buf_name);
            _LineMarkerRemoveType(wid, VCppData.gpictype_break_enabled);
            _LineMarkerRemoveType(wid, VCppData.gpictype_break_disabled);
            //_LineMarkerRemoveAllType(VCppData.gpictype_break_enabled);
            //_LineMarkerRemoveAllType(VCppData.gpictype_break_disabled);
            p_window_id.refresh();
            refresh();
         }
         return;
      }
      p_window_id = wid;
      int brkCounter = _dll_Brk_GetCnt();
      long brkPos[];
      int brkStatus[];
      int i;
      my_say("brkCounter = "brkCounter);
      if (brkCounter > 0){
         _dll_Brk_ResetEvalNdx();
         if (brkCounter >= 1024) {
            brkCounter = 0;
         }
         for (i = 0; i < brkCounter; i++) {
            brkPos[i] = _dll_Brk_GetPos();
            brkStatus[i] = _dll_Brk_GetState();
            _dll_Brk_IncEvalNdx();
            my_say(i': brkPos = 'brkPos[i]' brkStatus = 'brkStatus[i]);
         }
         _dll_Brk_ResetEvalNdx();
      }
      if (my_VS_UpdateUsedBrkBitmaps()) {
         //popup_message/*message*/("Clearing all breakpoint markers..." wid.p_buf_name);
         _LineMarkerRemoveType(wid, VCppData.gpictype_break_enabled);
         _LineMarkerRemoveType(wid, VCppData.gpictype_break_disabled);
         //_LineMarkerRemoveAllType(VCppData.gpictype_break_enabled);
         //_LineMarkerRemoveAllType(VCppData.gpictype_break_disabled);
         my_say("Restore breaks to "wid.p_buf_name" "_mdi.p_child.p_window_id.p_buf_name);
         for (i = 0; i < brkCounter; i++) {
            // drop pic here
            if (brkStatus[i]) {
               _LineMarkerAdd(wid, brkPos[i], 1, 1, VCppData._pic_break_enabled_marker, VCppData.gpictype_break_enabled, "");
            }
            else{
               _LineMarkerAdd(wid, brkPos[i], 1, 1, VCppData._pic_break_disabled_marker, VCppData.gpictype_break_disabled, "");
            }
         }
         p_window_id.refresh();
         refresh();
      }
   }
}

#if 0
// this is only called when come from VS here to SE, and for this
// we need a delay to retrieve the breakpoints, else VS will not answer
void my_VS_UpdateBreaksDelayed()
{
   my_say("my_VS_UpdateBreaksDelayed");
   if (VCppData.gMyUpdateBreaksTimerHandler > -1) {
      _kill_timer(VCppData.gMyUpdateBreaksTimerHandler);
      my_say("my_VS_UpdateBreaksDelayed _kill_timer(VCppData.gMyUpdateBreaksTimerHandler["VCppData.gMyUpdateBreaksTimerHandler"])");
      VCppData.gMyUpdateBreaksTimerHandler = -1;
      _post_call(my_VS_UpdateBreaks);
   }
}
#endif

// toggle breakpoint at current line
_command void my_VS_ToggleBreak() name_info (','VSARG2_MACRO|VSARG2_READ_ONLY/*|VSARG2_REQUIRES_EDITORCTL*/)
{
   // TODO: see debug_gui_update_breakpoints about how to use "_tbdebug_breakpoints_form" + "ctl_breakpoints_tree"

   //popup_message/*my_say*/("my_VS_ToggleBreak");
   if ( !_no_child_windows() ) {
      _dll_ToggleBreakpoint();
      _post_call(my_VS_UpdateBreaks);
   }
}


// ask MSDN help
_command void my_VS_Help() name_info (','VSARG2_MACRO|VSARG2_READ_ONLY/*|VSARG2_REQUIRES_EDITORCTL*/)
{
   //popup_message/*my_say*/("my_VS_Help");
   if ( !_no_child_windows() ) {
      _dll_SetVsWhereSeIs();
      _dll_VsHelp();
   }
}


// clear all VS breakpoints
_command void my_VS_ClearAllBreaks() name_info (','VSARG2_MACRO|VSARG2_READ_ONLY|VSARG2_REQUIRES_EDITORCTL)
{
   //popup_message/*my_say*/("my_VS_ClearAllBreaks");
   _dll_ClearBreakpoints();
   if ( !_no_child_windows() ) {
      _post_call(my_VS_UpdateBreaks);
   }
}


// show help
_command void my_VS_Help_About() name_info(',')
{
   _str msg =    "In order to sync SlickEdit from Visual Studio go to Visual Studio's \"Tools":+
                 "->External Tools\" menu,\nthen add a new tool\n    <SlickEdit synchro>\n":+
                 "with the Command:\n    C:\\Program Files\\SlickEditV18.0.1\\win\\vs.exe\n":+
                 "and the Arguments:\n":+
                 "    $(ItemPath) \"-#goto-line $(CurLine)\" \"-#goto-col $(CurCol)\" \"-#maximize-window\" \"-#SetKickBack\" \n\n":+ 
                 "Also you can map my-VS-SyncToSlickEdit to a key for fast switch to VisualStudio.\n\n":+
                 "SEtoVS.dll is expected to be present into":+
                 " your \"SlickEditV18.0.1\\win\" folder!\nYou can load it by using dload ":+
                 "SEtoVS command.";
   popup_imessage(msg);
}

// set keys in SlickEdit
_command void my_VS_Help_SetSEKeys() name_info(',')
{
   int status = _message_box("Would you like to set keybindings to the main features (jump to VS, toggle breakpoint)?\n":+
                             "If not, you can do/change it later from Tools->Options->KeyBindings and type my-VS- into commands box,\n":+
                             "or you can type my-VS-Help-SetSEKeys on Slick Edit command line, to start this dialog again.",
                       "VC++Sync.e",
                       MB_YESNO | MB_ICONQUESTION);
   if (status == IDYES) {
      gui_bind_to_key("my-VS-");
   }
}

// set ExternalTool in Visual Studio
_command void my_VS_Help_SetExternalToolIntoVS() name_info(',')
{
   _str msg =    "In order to sync SlickEdit from Visual Studio go to Visual Studio's \"Tools":+
                 "->External Tools\" menu,\nthen add a new tool\n    <SlickEdit synchro>\n":+
                 "with the Command:\n    C:\\Program Files\\SlickEditV18.0.1\\win\\vs.exe\n":+
                 "and the Arguments:\n":+
                 "    $(ItemPath) \"-#goto-line $(CurLine)\" \"-#goto-col $(CurCol)\" \"-#maximize-window\" \"-#SetKickBack\" \n\n":+ 
                 "Do you want to do it automatically now? (AddVSExternalTool.exe is expected in current path)\n":+
                 "If not, you can see later this dialog by typing my-VS-Help-SetExternalToolIntoVS on Slick Edit command line";
   int status = _message_box(msg,
                       "VC++Sync.e",
                       MB_YESNO | MB_ICONQUESTION);
   if (status == IDYES) {
      // compute executable path
      _str root_dir=get_env('VSROOT');
      _maybe_append_filesep(root_dir);
      _str myexe = root_dir :+ "win" :+ FILESEP :+ "AddVSExternalTool.exe";
      execute (maybe_quote_filename(myexe), 'W'); // wait for termination
  }
}

// init globals (at module loading/initialisation)
void init_all()
{
   //VCppData.gMyUpdateBreaksTimerHandler = -1;
   //VCppData.gTempDisableSwitchBuffUpdate = false;
   VCppData.gOldBuffName = "";
   VCppData._pic_break_enabled_marker = -1;
   VCppData.gpictype_break_enabled = -1;
   VCppData._pic_break_disabled_marker = -1;
   VCppData.gpictype_break_disabled = -1;
}


// http://community.slickedit.com/index.php?topic=1907.0
// Global variable set in main.e AFTER all definit()s have run
_str _editor_cmdline;
static void Maybe_init_all()
{
   // Only show splash if the the -mdihide switch was not passed to this instance of the application
   if( 0 == pos("-mdihide",_editor_cmdline,1,'i') ) {
      init_all();
   }
}
/*
    The Slick-C language provides two module initialization functions
    called definit and defload. If the two are present, the procedures
    definit and defload are called when a module is loaded. The definit
    module is called before defload. When the module is saved by the
    write_state command, the definit procedure is invoked each time the
    editor is invoked. This gives your module an opportunity to perform
    initializations such as creating a temporary file, or allocating a
    selection, or bookmark
*/
definit()
{
   // this is called everytime SlickEdit starts!
   my_say("\"VC++Sync.e\" definit");
   if (arg(1)!='L') {
      _post_call(Maybe_init_all);
   }
}



// when load the module (menu Macro -> Load Module ...)
defload()
{
   // this is called only when you load the module for the first time!
   _str msg =    "You can see again some instructions by typing my-VS-Help-About on the SlickEdit's command line!\n\n\n\n":+
                 "SEtoVS.dll is expected to be present into":+
                 " your \"SlickEditV18.0.1\\win\" folder!\nYou can load it later by using dload ":+
                 "SEtoVS inte SE's command line.\n\nDo you want to load it now?";
   int status = _message_box(msg,
                       "VC++Sync.e",
                       MB_YESNO | MB_ICONQUESTION);
   if (status == IDYES) {
      if (my_VS_LoadDll()) {
         //popup_message("Please restart SlickEdit!", MB_ICONEXCLAMATION, "VC++Sync.e");
      }
   }
   init_all();
   my_VS_Help_SetExternalToolIntoVS();
   my_VS_Help_SetSEKeys();
}

// called when Slick calls it's list of _switchbuf_ - fct.s
// THIS IS HAPPENING WHEN CURRENT FILE IS CHANGED!
// update the breakpoints from Visual Studio
void _switchbuf_VCppSync (_str oldbuffname, _str flag)
{
   if (! gIsDllLoaded) {
      return;
   }
   /*if (VCppData.gTempDisableSwitchBuffUpdate) {
      my_say("_switchbuf_VCppSync exits because VCppData.gTempDisableSwitchBuffUpdate is set!");
      return;
   }*/
   if (!_no_child_windows() && (oldbuffname == "")) {
      _str newbuffname = maybe_quote_filename(_mdi.p_child.p_buf_name);
      if(lowcase(VCppData.gOldBuffName) != lowcase(newbuffname))
      {
         VCppData.gOldBuffName = newbuffname;
         my_say('_switchbuf_VCppSync 'VCppData.gOldBuffName);
         _post_call(my_VS_UpdateBreaks);
      }
   }
}

// called when Slick calls it's list of _actapp_ - fct.s
// THIS IS HAPPENING WHEN SE GETS FOCUS!
// update the breakpoints from Visual Studio
void _actapp_VCppSync(typeless gettingFocus='')
{
      // IF application got focus
   if (gettingFocus) {
      if (! gIsDllLoaded) {
         return;
      }
      /*if (VCppData.gTempDisableSwitchBuffUpdate) {
         my_say("_switchbuf_VCppSync exits because VCppData.gTempDisableSwitchBuffUpdate is set!");
         return;
      }*/
      if (!_no_child_windows()) {
         _str buffname = maybe_quote_filename(_mdi.p_child.p_buf_name);
         if(length(buffname) > 0)
         {
            my_say('_actapp_VCppSync 'buffname);
            _post_call(my_VS_UpdateBreaks);
         }
      }
   }
}


_command int SetKickBack() name_info(','VSARG2_CMDLINE|VSARG2_MARK|VSARG2_READ_ONLY)
{
   _dll_SetKickBack(); // extract breakpoints using VS's current document, not SE's current buffer
   _post_call(my_VS_UpdateBreaks);
   return 0;
}


